import argparse
import os.path
from copy import deepcopy
from typing import Dict, Optional, Union

import cv2
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from simba.mixins.abstract_classes import AbstractFeatureExtraction
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.image_mixin import ImageMixin
from simba.plotting.geometry_plotter import GeometryPlotter
from simba.utils.checks import check_instance, check_int
from simba.utils.enums import Formats
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_video_meta_data, read_df,
                                    read_frm_of_video)

dpi = matplotlib.rcParams["figure.dpi"]

CONFIG_PATH = "/Users/simon/Desktop/envs/simba/troubleshooting/RI/project_folder/project_config.ini"
VIDEO_NAME = "RI_01_165_clipped"
WHITE = "Animal_1"
BLACK = "Animal_2"

config = ConfigReader(config_path=CONFIG_PATH, create_logger=False)
data_path = os.path.join(
    config.outlier_corrected_dir, f"{VIDEO_NAME}.{config.file_type}"
)
video_path = os.path.join(config.video_dir, f"{VIDEO_NAME}.mp4")
get_video_meta_data(video_path)
df = read_df(file_path=data_path, file_type=config.file_type)
resolution, pixels_per_mm, fps = config.read_video_info(video_name=VIDEO_NAME)

white_animal_bp_names, black_animal_bp_names = (
    config.animal_bp_dict[WHITE],
    config.animal_bp_dict[BLACK],
)
white_animal_cols, black_animal_cols = [], []
for x, y in zip(white_animal_bp_names["X_bps"], white_animal_bp_names["Y_bps"]):
    white_animal_cols.extend((x, y))
for x, y in zip(black_animal_bp_names["X_bps"], black_animal_bp_names["Y_bps"]):
    black_animal_cols.extend((x, y))
white_animal_df, black_animal_df = df[white_animal_cols].astype(int), df[
    black_animal_cols
].astype(int)
white_animal_df_arr = white_animal_df.values.reshape(len(white_animal_df), -1, 2)
black_animal_df_arr = black_animal_df.values.reshape(len(black_animal_df), -1, 2)

white_animal_polygons = GeometryMixin().multiframe_bodyparts_to_polygon(
    data=white_animal_df_arr,
    pixels_per_mm=pixels_per_mm,
    parallel_offset=40,
    verbose=False,
    core_cnt=-1,
)
black_animal_polygons = GeometryMixin().multiframe_bodyparts_to_polygon(
    data=black_animal_df_arr,
    pixels_per_mm=pixels_per_mm,
    parallel_offset=40,
    verbose=False,
    core_cnt=-1,
)


black_animal_center = df[["Nose_2_x", "Nose_2_y"]].values.astype(np.int)
black_animal_circles = GeometryMixin().multiframe_bodyparts_to_circle(
    data=black_animal_center,
    pixels_per_mm=pixels_per_mm,
    parallel_offset=100,
    core_cnt=-1,
)


# white_animal_polygons = GeometryMixin().multiframe_minimum_rotated_rectangle(
#     shapes=white_animal_polygons, verbose=False, core_cnt=-1
# )
# black_animal_polygons = GeometryMixin().multiframe_minimum_rotated_rectangle(
#     shapes=black_animal_polygons, verbose=False, core_cnt=-1
# )

# IF WE WANT, THEN WE CAN VISUALIZE THE ANIMAL GEOMETRIES.
imgs = ImageMixin().slice_shapes_in_imgs(
    imgs=video_path, shapes=white_animal_polygons, verbose=False
)
ImageMixin.img_stack_to_video(
    imgs=imgs,
    save_path="/Users/simon/Desktop/envs/simba/troubleshooting/RI/project_folder/frames/output/stack/test_3.mp4",
    fps=fps,
)

im = imgs[4]
height, width, depth = im.shape
figsize = width / float(dpi), height / float(dpi)

plt.figure(figsize=figsize)
plt.axis("off")
plt.imshow(im)
plt.show()


# Entire_video
GeometryPlotter(
    config_path=CONFIG_PATH,
    geometries=[white_animal_polygons, black_animal_polygons],
    video_name=VIDEO_NAME,
    core_cnt=-1,
    save_path="/Users/simon/Desktop/envs/simba/troubleshooting/RI/project_folder/frames/output/stack/test_2.mp4",
).run()

# #OVERLAP
# overlap = GeometryMixin().multiframe_compute_pct_shape_overlap(shape_1=white_animal_polygons, shape_2=black_animal_polygons, verbose=False)

# #area
# white_animal_area = GeometryMixin().multiframe_area(shapes=white_animal_polygons, pixels_per_mm=pixels_per_mm, verbose=False)
#
# #skeleton
# white_animal_skeleton = [['Ear_left_1', 'Nose_1'],
#                          ['Ear_left_1', 'Ear_right_1'],
#                          ['Ear_right_1', 'Nose_1'],
#                          ['Nose_1', 'Center_1'],
#                          ['Center_1', 'Lat_left_1'],
#                          ['Center_1', 'Lat_right_1'],
#                          ['Center_1', 'Tail_base_1']]
#
# black_animal_skeleton = [['Ear_left_2', 'Nose_2'],
#                          ['Ear_left_2', 'Ear_right_2'],
#                          ['Ear_right_2', 'Nose_2'],
#                          ['Nose_2', 'Center_2'],
#                          ['Center_2', 'Lat_left_2'],
#                          ['Center_2', 'Lat_right_2'],
#                          ['Center_2', 'Tail_base_2']]
#
# white_animal_skeleton = GeometryMixin().multiframe_bodyparts_to_multistring_skeleton(data_df=df, skeleton=white_animal_skeleton, verbose=False)
# black_animal_skeleton = GeometryMixin().multiframe_bodyparts_to_multistring_skeleton(data_df=df, skeleton=black_animal_skeleton, verbose=False)


# union
# bodyparts_to_multistring_skeleton



__author__ = "Tzuk Polinsky"

import os
import random

import cv2
import numpy as np
import pandas as pd

from simba.data_processors.directing_animal_to_bodypart import \
    DirectingAnimalsToBodyPartAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.data import create_color_palettes
from simba.utils.enums import Formats
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df
from simba.utils.warnings import NoDataFoundWarning


class DirectingAnimalsToBodyPartVisualizer(ConfigReader, PlottingMixin):
    """
    Visualize when animals are directing towards their own body-parts .

    .. important::
       Requires the pose-estimation data for the left ear, right ears and nose of individual animals.
       For better runtime, use :meth:`simba.plotting.Directing_animals_visualizer.DirectingOtherAnimalsVisualizerMultiprocess`.

    .. note::
       `Example of expected output <https://www.youtube.com/watch?v=d6pAatreb1E&list=PLi5Vwf0hhy1R6NDQJ3U28MOUJPfl2YWYl&index=22`_.

        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>_.

        .. image:: _static/img/directing_other_animals.png
           :width: 450
           :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_path: path to data file
    :parameter dict style_attr: Visualisation attributes (colors and sizes etc.)


    Examples
    -----
    >>> style_attr = {'Show_pose': True,
    >>>               'Pose_circle_size': 3,
    >>>               "Direction_color": 'Random',
    >>>               'Direction_thickness': 4, 'Highlight_endpoints': True}
    >>> _ = DirectingAnimalsToBodyPartVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini', video_name='Testing_Video_3.mp4', style_attr=style_attr).run()
    """

    def __init__(self, config_path: str, data_path: str, style_attr: dict):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)

        self.data_path = data_path
        _, self.video_name, _ = get_fn_ext(self.data_path)
        self.direction_analyzer = DirectingAnimalsToBodyPartAnalyzer(
            config_path=config_path,
        )
        self.data_dict = self.direction_analyzer.read_directionality_dfs()
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.style_attr, self.pose_colors = style_attr, []
        self.colors = get_color_dict()
        if self.style_attr["Show_pose"]:
            self.pose_colors = create_color_palettes(
                self.animal_cnt, int(len(self.x_cols) + 1)
            )
        if self.style_attr["Direction_color"] == "Random":
            self.direction_colors = create_color_palettes(1, int(self.animal_cnt**2))
        else:
            self.direction_colors = [self.colors[self.style_attr["Direction_color"]]]
        self.video_path = self.find_video_of_file(
            video_dir=self.video_dir, filename=self.video_name
        )
        if not os.path.exists(self.directing_body_part_animal_video_output_path):
            os.makedirs(self.directing_body_part_animal_video_output_path)
        print(f"Processing video {self.video_name}...")

    def run(self):
        """
        Method to create directionality videos. Results are stored in
        the `project_folder/frames/output/ROI_directionality_visualize` directory of the SimBA project

        Returns
        ----------
        None
        """

        self.data_df = read_df(self.data_path, file_type=self.file_type)
        self.video_save_path = os.path.join(
            self.directing_body_part_animal_video_output_path, self.video_name + ".mp4"
        )
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.video_data = self.data_dict[self.video_name]
        self.writer = cv2.VideoWriter(
            self.video_save_path,
            self.fourcc,
            self.video_meta_data["fps"],
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        self.__create_video()
        self.timer.stop_timer()

    def __draw_individual_lines(self, animal_img_data: pd.DataFrame, img: np.array):
        color = self.direction_colors[0]
        bp_x_name = "Animal_" + self.bodypart_direction + "_x"
        bp_y_name = "Animal_" + self.bodypart_direction + "_y"
        for cnt, (i, r) in enumerate(animal_img_data.iterrows()):
            if self.style_attr["Direction_color"] == "Random":
                color = random.sample(self.direction_colors[0], 1)[0]
            if r["Directing_BOOL"]:
                cv2.line(
                    img,
                    (int(r["Eye_x"]), int(r["Eye_y"])),
                    (int(r[bp_x_name]), int(r[bp_y_name])),
                    color,
                    self.style_attr["Direction_thickness"],
                )

        return img

    def __create_video(self):
        img_cnt = 0
        video_timer = SimbaTimer(start=True)
        video_timer.start_timer()
        color = self.direction_colors[0]
        frames_index = dict(enumerate(self.video_data["Frame_#"].unique()))

        while self.cap.isOpened():
            ret, img = self.cap.read()
            try:
                if ret:
                    if self.style_attr["Show_pose"]:
                        bp_data = self.data_df.iloc[img_cnt]
                        for cnt, (animal_name, animal_bps) in enumerate(
                            self.animal_bp_dict.items()
                        ):
                            for bp in zip(animal_bps["X_bps"], animal_bps["Y_bps"]):
                                x_bp, y_bp = bp_data[bp[0]], bp_data[bp[1]]
                                cv2.circle(
                                    img,
                                    (int(x_bp), int(y_bp)),
                                    self.style_attr["Pose_circle_size"],
                                    self.animal_bp_dict[animal_name]["colors"][cnt],
                                    self.style_attr["Direction_thickness"],
                                )

                    if img_cnt in frames_index:
                        img_data = self.video_data[
                            self.video_data["Frame_#"] == img_cnt
                        ]
                        if self.style_attr["Polyfill"]:
                            convex_hull_arr = img_data.loc[
                                0, ["Eye_x", "Eye_y"]
                            ].values.reshape(-1, 2)
                        else:
                            img = self.__draw_individual_lines(
                                animal_img_data=img_data, img=img
                            )

                    img_cnt += 1
                    self.writer.write(np.uint8(img))
                    # print(
                    #     "Frame: {} / {}. Video: {}".format(
                    #         str(img_cnt),
                    #         str(self.video_meta_data["frame_count"]),
                    #         self.video_name,
                    #     )
                    # )
                else:
                    self.cap.release()
                    self.writer.release()
                    break

            except IndexError:
                self.cap.release()
                self.writer.release()

        video_timer.stop_timer()
        stdout_success(
            msg=f"Directionality video {self.video_name} saved in {self.directing_body_part_animal_video_output_path} directory",
            elapsed_time=video_timer.elapsed_time_str,
        )



__author__ = "Simon Nilsson"

try:
    from cuml.cluster import hdbscan
    from cuml.cluster.hdbscan import HDBSCAN

    gpu_flag = True
except ModuleNotFoundError:
    from hdbscan import HDBSCAN
    import hdbscan

import glob
import itertools
import os
import random

import pandas as pd

from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.unsupervised.umap_embedder import UmapEmbedder
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.printing import SimbaTimer, stdout_success


class HDBSCANClusterer(UnsupervisedMixin):
    """
    Methods for grid-search HDBSCAN model fit and transform.
    Defaults to GPU and cuml.cluster.HDBSCAN. If GPU unavailable, then hdbscan.HDBSCAN.
    """

    def __init__(self):
        super().__init__()

    def fit(self, data_path: str, save_dir: str, hyper_parameters: dict):
        """
        :param data_path: Path holding pickled unsupervised dimensionality reduction results in ``data_map.yaml`` format
        :param save_dir: Empty directory where to save the HDBSCAN results.
        :param hyper_parameters: dict holding hyperparameters in list format
        :return:

        :Example I: Grid-search fit:
        >>> hyper_parameters = {'alpha': [1.0], 'min_cluster_size': [10], 'min_samples': [1], 'cluster_selection_epsilon': [20]}
        >>> embedding_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models'
        >>> save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models'
        >>> config_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini'
        >>> clusterer = HDBSCANClusterer(data_path=embedding_dir, save_dir=save_dir)
        >>> clusterer.fit(hyper_parameters=hyper_parameters)
        """

        self.save_dir, self.data_path = save_dir, data_path
        self.check_that_directory_is_empty(directory=self.save_dir)
        if os.path.isdir(data_path):
            check_if_dir_exists(in_dir=data_path)
            check_if_filepath_list_is_empty(
                filepaths=glob.glob(data_path + "/*.pickle"),
                error_msg=f"SIMBA ERROR: No pickle files in {data_path}",
            )
        else:
            check_file_exist_and_readable(file_path=data_path)
            self.data_path = data_path

        self.search_space = list(
            itertools.product(
                *[
                    hyper_parameters[Clustering.ALPHA.value],
                    hyper_parameters[Clustering.MIN_CLUSTER_SIZE.value],
                    hyper_parameters[Clustering.MIN_SAMPLES.value],
                    hyper_parameters[Clustering.EPSILON.value],
                ]
            )
        )
        self.embeddings = self.read_pickle(data_path=self.data_path)
        print(
            f"Fitting {str(len(self.search_space) * len(self.embeddings.keys()))} HDBSCAN model(s)..."
        )
        self.__fit_hdbscan()
        self.timer.stop_timer()
        stdout_success(
            msg=f"{str(len(self.search_space) * len(self.embeddings.keys()))} saved in {self.save_dir}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __fit_hdbscan(self):
        self.model_counter = 0
        for k, v in self.embeddings.items():
            fit_timer = SimbaTimer()
            fit_timer.start_timer()
            embedder = v[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value]
            for cnt, h in enumerate(self.search_space):
                results, self.model = {}, {}
                self.model_counter += 1
                self.model_timer = SimbaTimer()
                self.model_timer.start_timer()
                self.model[Unsupervised.HASHED_NAME.value] = random.sample(
                    self.model_names, 1
                )[0]
                self.model[Unsupervised.PARAMETERS.value] = {
                    Clustering.ALPHA.value: h[0],
                    Clustering.MIN_CLUSTER_SIZE.value: h[1],
                    Clustering.MIN_SAMPLES.value: h[2],
                    Clustering.EPSILON.value: h[3],
                }
                self.model[Unsupervised.MODEL.value] = HDBSCAN(
                    algorithm="best",
                    alpha=self.model[Unsupervised.PARAMETERS.value][
                        Clustering.ALPHA.value
                    ],
                    approx_min_span_tree=True,
                    gen_min_span_tree=True,
                    min_cluster_size=self.model[Unsupervised.PARAMETERS.value][
                        Clustering.MIN_CLUSTER_SIZE.value
                    ],
                    min_samples=self.model[Unsupervised.PARAMETERS.value][
                        Clustering.MIN_SAMPLES.value
                    ],
                    cluster_selection_epsilon=self.model[Unsupervised.PARAMETERS.value][
                        Clustering.EPSILON.value
                    ],
                    p=None,
                    prediction_data=True,
                )

                self.model[Unsupervised.MODEL.value].fit(embedder.embedding_)
                results[Unsupervised.DATA.value] = v[Unsupervised.DATA.value]
                results[Unsupervised.METHODS.value] = v[Unsupervised.METHODS.value]
                results[Unsupervised.DR_MODEL.value] = v[Unsupervised.DR_MODEL.value]
                results[Clustering.CLUSTER_MODEL.value] = self.model
                self.__save(data=results)

    def __save(self, data: dict) -> None:
        self.write_pickle(
            data=data,
            save_path=os.path.join(
                self.save_dir, f"{self.model[Unsupervised.HASHED_NAME.value]}.pickle"
            ),
        )
        self.model_timer.stop_timer()
        stdout_success(
            msg=f"Model {self.model_counter}/{len(self.search_space) * len(list(self.embeddings.keys()))} ({self.model[Unsupervised.HASHED_NAME.value]}) saved...",
            elapsed_time=self.model_timer.elapsed_time,
        )

    def transform(
        self,
        data_path: str,
        model: str or dict,
        save_dir: str or None = None,
        settings: dict or None = None,
    ):
        """
        :param data_path: Path to directory holding pickled unsupervised dimensionality reduction results in ``data_map.yaml`` format
        :param model: Path to pickle holding hdbscan model in ``data_map.yaml`` format.
        :param save_dir: Empty directory where to save the HDBSCAN results. If none, then keep results in memory under self.results.
        :param settings: User-defined params.

        :Example I: Transform:
        >>> data_path = 'project_folder/logs/unsupervised_data_20230416145821.pickle'
        >>> save_path = 'unsupervised/dr_models'
        >>> clusterer = HDBSCANClusterer(data_path=data_path, save_dir=save_path)
        >>> clusterer.transform(model='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/awesome_curran.pickle', settings={'DATA': None}, data_path=data_path)
        """

        timer = SimbaTimer()
        timer.start_timer()
        if isinstance(model, str):
            check_file_exist_and_readable(file_path=model)
            model = self.read_pickle(data_path=model)

        check_file_exist_and_readable(file_path=data_path)
        umap_embedder = UmapEmbedder()
        umap_embedder.transform(
            model=model, data_path=data_path, settings={"DATA": None, "format": None}
        )
        self.label, self.strength = hdbscan.approximate_predict(
            model[Clustering.CLUSTER_MODEL.value][Unsupervised.MODEL.value],
            umap_embedder.results[["X", "Y"]].values,
        )
        self.results = pd.DataFrame(
            {"HDBSCAN_LABEL": self.label, "HDBSCAN_STRENGTH": self.strength},
            columns=["HDBSCAN_LABEL", "HDBSCAN_STRENGTH"],
            index=umap_embedder.umap_df.index,
        )
        self.results = pd.concat(
            [umap_embedder.results[["X", "Y"]], self.results], axis=1
        )
        if settings[Unsupervised.DATA.value] == Unsupervised.SCALED.value:
            self.results = pd.concat(
                [umap_embedder.scaled_umap_data, self.results], axis=1
            )
        elif settings[Unsupervised.DATA.value] == Unsupervised.RAW.value:
            self.results = pd.concat([umap_embedder.umap_df, self.results], axis=1)
        if save_dir:
            save_path = os.path.join(
                save_dir,
                f"Transformed_{model[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
            )
            if settings[Unsupervised.FORMAT.value] is Unsupervised.CSV.value:
                self.results.to_csv(save_path)
            timer.stop_timer()
            print(
                f"Transformed data saved at {save_dir} (elapsed time: {timer.elapsed_time_str}s)"
            )


# hyper_parameters = {'alpha': [1.0], 'min_cluster_size': [10], 'min_samples': [1], 'cluster_selection_epsilon': [20]}
# embedding_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models'
# config_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini'
# clusterer = HDBSCANClusterer(data_path=embedding_dir, save_dir=save_dir)
# clusterer.fit(hyper_parameters=hyper_parameters)


# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
# save_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models'
# clusterer = HDBSCANClusterer(data_path=data_path, save_dir=save_path)
# clusterer.transform(model='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/awesome_curran.pickle', settings={'DATA': None}, data_path=data_path)

# settings = {'feature_values': True, 'scaled_features': True, 'save_format': 'csv'}
# clusterer_model_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/amazing_burnell.pickle'
# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230215093552.pickle'
# save_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models'

# _ = HDBSCANTransform(clusterer_model_path=clusterer_model_path,
#                      data_path=data_path,
#                      save_dir=save_path,
#                      settings=settings)



import json
import math
import os
import random
import statistics
import sys

import dash
import dash_color_picker as dcp
import dash_colorscales as dcs
import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
import h5py
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash.dependencies import Input, Output, State

filePath = "/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/SimBA_dash_storage_20230202105812.h5"
groupPath = "None"

# if not len(sys.argv) > 1:
#     print("Expecting link argument.")
# else:
#     print("in p1.py link is " + sys.argv[1] + ' ' + sys.argv[2])
#     filePath = sys.argv[1]
#     groupPath = sys.argv[2]


external_stylesheets = [
    "dash_simba_base.css",
    "//use.fontawesome.com/releases/v5.0.7/css/all.css",
]
app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
hf = h5py.File(filePath, "r")


# Input: string path of group
# Output: returns list of groups, dictionary of group and the videos in each group, and if groups exist
def get_groups(path):
    g_list = []
    g_dict = {}

    is_file = os.path.isfile(path)
    if is_file:
        df_groups = pd.read_csv(path)
        for group in df_groups.keys():
            g_list.append(group)
            g_dict[group] = df_groups[group].dropna().tolist()
    else:
        pass
    return g_list, g_dict, is_file


# Input: List of videos, behaviors, categories, hdf5 file path, and hdf5 file object
# Output: Returns complete dictionary of all behaviors, categories, and features
def get_feature_data(videos, behaviors, g_categories, path, h5file):
    dict_features = {}

    for b in behaviors:
        dict_features[b] = {}

        for category in g_categories:
            if category == "VideoData":
                print(videos)
                video = pd.read_hdf(path, key=category + "/" + videos[0])
                dict_features[b][category] = {
                    column: get_probability_data(column, videos, path)
                    for column in video.columns
                    if "_" + b in column
                }
            else:
                for sub_c in h5file.get(category).keys():
                    video = pd.read_hdf(path, key=category + "/" + sub_c)

                    if b == "Attack":
                        dict_features[b][sub_c] = {
                            column: video[column] for column in video.columns
                        }
                    else:
                        dict_features[b][sub_c] = {
                            column: video[column]
                            for column in video.columns
                            if b in column
                        }

    return dict_features


# Input: Current feature name, list of videos, and path of hdf5 file
# Output: Returns pandas dataframe for each video with probability features
def get_probability_data(feature, list_videos, path_name):
    feature_data = pd.DataFrame()

    for video in list_videos:
        read_video = pd.read_hdf(path_name, key="VideoData/" + video)
        feature_data[video] = read_video[feature]
        # dict_data[f][video].fillna(0)

    for g in groups_dict:
        feature_data[g + "_Mean"] = feature_data[groups_dict[g]].mean(axis=1)

    mean_data = feature_data.mean(axis=1)
    feature_data["Total_Mean"] = mean_data

    return feature_data


# Input: hdf5 file object
# Output: Returns list of categories and list of all hdf5 main keys
def get_categories(h5file):
    categories = []
    general_list = []
    for k in h5file.keys():
        if k != "Classifier_names" and k != "Video_info":
            if "VideoData" in k:
                categories.append(k)
                print("VideoData categories", k)
            else:
                hf_category = h5file.get(k)
                for key in hf_category.keys():
                    categories.append(key)
                    print("VideoData categories 1", k)
            general_list.append(k)
    return categories, general_list


# Input: Boolean if groups exist
# Output: Returns either visible or hidden style
def set_display(exist):
    if exist:
        return {"display": "block"}
    else:
        return {"display": "none"}


# Return download path
def get_download_path():
    """Returns the default downloads path for linux or windows"""
    if os.name == "nt":
        import winreg

        sub_key = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
        downloads_guid = "{374DE290-123F-4565-9164-39C4925E467B}"
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
            location = winreg.QueryValueEx(key, downloads_guid)[0]
        return location
    else:
        return os.path.join(os.path.expanduser("~"), "downloads")


# # VARIABLES AND DATA
# # List of all videos in H5File
video_info = pd.read_hdf(filePath, key="Video_info")
H5list_videos = list(video_info["Video"])

# # List of Behaviors
classifiers = pd.read_hdf(filePath, key="Classifier_names")
list_behaviors = list(classifiers["Classifier_names"])

# # List of Categories
list_categories, general_categories = get_categories(hf)

# # List of Groups and which videos belong in which group
group_list, groups_dict, groups_exist = get_groups(groupPath)

# # Dictionary of fps for each video
fps_dict = dict(zip(H5list_videos, video_info["fps"]))
for group in groups_dict:
    try:
        curr_fps = fps_dict[groups_dict[group][0]]
        fps_dict[group] = curr_fps
        for video in groups_dict[group]:
            if fps_dict[video] != curr_fps:
                print("Videos do not have the same fps")
    except IndexError:
        pass

# # Create and store all the data to be used in dash plotly in the form of a dictionary
all_data_behaviors = get_feature_data(
    H5list_videos, list_behaviors, general_categories, filePath, hf
)


# # List of available and supported fonts
FONTS = [
    "Verdana",
    "Helvetica",
    "Calibri",
    "Arial",
    "Arial Narrow",
    "Candara",
    "Geneva",
    "Courier New",
    "Times New Roman",
]

# # Default graph configuration
config = {
    "toImageButtonOptions": {
        "format": "svg",  # one of png, svg, jpeg, webp
        "filename": "custom_image",
        "height": 500,
        "width": 700,
        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
    }
}

hf.close()


# Returns color pickers based on number of groups
def create_color_pickers():
    output = []
    for g in group_list:
        new_picker = [
            html.Label(g),
            dcp.ColorPicker(id="ColorPicker" + str(g), color="#609BE0"),
        ]
        output += new_picker
    output += [html.Label("Mean"), dcp.ColorPicker(id="ColorPickerMean", color="#f22")]
    return output


app.layout = html.Div(
    children=[
        html.Div(
            id="header",
            className="row",
            children=[
                html.H1(
                    children="SimBA Interactive Data Visualization Dashboard",
                    className="nine columns",
                ),
                html.Img(
                    src=app.get_asset_url("TheGoldenLab.PNG"),
                    className="three columns",
                    style={
                        "height": "15%",
                        "width": "15%",
                        "float": "right",
                        "position": "relative",
                    },
                ),
            ],
        ),
        html.Div(
            id="simba-control-tabs",
            className="three columns",
            children=[
                dcc.Tabs(
                    id="simba-tabs",
                    value="all_data",
                    children=[
                        dcc.Tab(
                            label="Data",
                            value="all_data",
                            children=[
                                html.Div(
                                    id="plot_data",
                                    children=[
                                        html.H6("Select Behavior to Plot"),
                                        html.Label("Behaviors:"),
                                        dcc.Dropdown(
                                            # className='eight columns',
                                            id="behaviors",
                                            options=[
                                                {"label": b, "value": b}
                                                for b in list_behaviors
                                            ],
                                            multi=False,
                                            value=list_behaviors[0],
                                            clearable=False,
                                        ),
                                        html.Label("Category:"),
                                        dcc.Dropdown(
                                            id="category_selector",
                                            # className='eight columns',
                                            clearable=False,
                                        ),
                                        html.Label("Feature:"),
                                        dcc.Dropdown(
                                            id="feature_selector", clearable=False
                                        ),
                                        html.Div(
                                            id="multi_group_data",
                                            children=[
                                                html.Hr(),
                                                html.H6("Plotting Group Means"),
                                                html.Label("Select Group(s):"),
                                                dcc.Dropdown(
                                                    id="multi_group_selector",
                                                    # className='eight columns',
                                                    options=[
                                                        {"label": key, "value": key}
                                                        for key in groups_dict.keys()
                                                    ],
                                                    multi=True,
                                                    clearable=False,
                                                    value=group_list,
                                                ),
                                                dcc.Checklist(
                                                    id="group_mean",
                                                    options=[
                                                        {
                                                            "label": "Show Total Mean",
                                                            "value": "showGroupMean",
                                                            "disabled": False,
                                                        }
                                                    ],
                                                    value=["showGroupMean"],
                                                    labelStyle={
                                                        "display": "inline-block"
                                                    },
                                                ),
                                            ],
                                            style=set_display(groups_exist),
                                        ),
                                        html.Hr(),
                                        html.H6("Plotting Individual Groups"),
                                        html.Div(
                                            id="show_group_opt",
                                            children=[
                                                html.Label(
                                                    "Select Single Group to Plot:",
                                                    style=set_display(groups_exist),
                                                ),
                                                dcc.Dropdown(
                                                    id="group_selector",
                                                    options=[
                                                        {"label": key, "value": key}
                                                        for key in groups_dict.keys()
                                                    ],
                                                    multi=False,
                                                    clearable=False,
                                                ),
                                            ],
                                            style=set_display(groups_exist),
                                        ),
                                        html.Label("Select Video(s):"),
                                        dcc.Dropdown(
                                            id="video_numbers",
                                            multi=True,
                                            clearable=False,
                                        ),
                                        dcc.Checklist(
                                            id="video_mean",
                                            options=[
                                                {
                                                    "label": "Show Group Mean",
                                                    "value": "showVideoMean",
                                                    "disabled": False,
                                                }
                                            ],
                                            value=["showVideoMean"],
                                        ),
                                    ],
                                )
                            ],
                        ),
                        dcc.Tab(
                            label="Graph Settings",
                            value="graph_settings",
                            children=[
                                html.Div(
                                    className="twelve columns",
                                    id="graph_settings",
                                    children=[
                                        html.Div(
                                            id="color_properties",
                                            className="twelve columns",
                                            children=[
                                                html.H6("Color Properties"),
                                                html.Div(
                                                    id="color_pickers",
                                                    children=create_color_pickers(),
                                                ),
                                                html.Label(
                                                    "Colorscale Gradient for Individual Videos"
                                                ),
                                                dcs.DashColorscales(
                                                    id="ind_group_colors",
                                                    colorscale=[
                                                        "#fbb4ae",
                                                        "#b3cde3",
                                                        "#ccebc5",
                                                        "#decbe4",
                                                        "#fed9a6",
                                                        "#ffffcc",
                                                    ],
                                                ),
                                                html.Button(
                                                    "Update Colors",
                                                    id="update_graph",
                                                    n_clicks=0,
                                                    type="submit",
                                                ),
                                                html.Hr(),
                                            ],
                                        ),
                                        html.Div(
                                            className="twelve columns",
                                            id="probability-graph-properties",
                                            children=[
                                                html.H6("Probability Graph Properties"),
                                                html.Div(
                                                    id="multi_prob_prop",
                                                    children=[
                                                        html.Label(
                                                            "Set Group Viewing Frames:"
                                                        ),
                                                        dcc.Input(
                                                            id="multi_start_x",
                                                            type="number",
                                                            className="four columns",
                                                        ),
                                                        dcc.Input(
                                                            id="multi_end_x",
                                                            type="number",
                                                            className="four columns",
                                                        ),
                                                        html.Button(
                                                            id="submit_multi_range",
                                                            className="three columns",
                                                            n_clicks=1,
                                                            children="Submit",
                                                            type="submit",
                                                        ),
                                                        html.Label(
                                                            "Multi Group (Seconds)"
                                                        ),
                                                        dcc.Input(
                                                            id="multi_start_sec",
                                                            type="number",
                                                            className="four columns",
                                                            disabled=True,
                                                        ),
                                                        dcc.Input(
                                                            id="multi_end_sec",
                                                            type="number",
                                                            className="four columns",
                                                            disabled=True,
                                                        ),
                                                        html.Button(
                                                            id="reset_multi_range",
                                                            className="twelve columns",
                                                            n_clicks=1,
                                                            children="Reset Axes",
                                                            type="reset",
                                                        ),
                                                    ],
                                                    style=set_display(groups_exist),
                                                ),
                                                html.Label(
                                                    "Set Video Viewing Frames:",
                                                    className="twelve columns",
                                                ),
                                                dcc.Input(
                                                    id="ind_start_x",
                                                    type="number",
                                                    className="four columns",
                                                ),
                                                dcc.Input(
                                                    id="ind_end_x",
                                                    type="number",
                                                    className="four columns",
                                                ),
                                                html.Button(
                                                    id="submit_ind_range",
                                                    className="three columns",
                                                    n_clicks=1,
                                                    children="Submit",
                                                    type="submit",
                                                ),
                                                html.Label(
                                                    "Individual Group (Seconds)",
                                                    className="twelve columns",
                                                ),
                                                dcc.Input(
                                                    id="ind_start_sec",
                                                    type="number",
                                                    className="four columns",
                                                    disabled=True,
                                                ),
                                                dcc.Input(
                                                    id="ind_end_sec",
                                                    type="number",
                                                    className="four columns",
                                                    disabled=True,
                                                ),
                                                html.Button(
                                                    id="reset_ind_range",
                                                    className="twelve columns",
                                                    n_clicks=1,
                                                    children="Reset Axes",
                                                    type="reset",
                                                ),
                                            ],
                                            style={"display": "block"},
                                        ),
                                        html.Div(
                                            className="twelve columns",
                                            id="bar-graph-properties",
                                            children=[
                                                html.H6("Bar Graph Properties"),
                                                html.Label("Error Type:"),
                                                dcc.RadioItems(
                                                    id="error_type",
                                                    options=[
                                                        {
                                                            "label": "Above",
                                                            "value": "above",
                                                        },
                                                        {
                                                            "label": "Below",
                                                            "value": "below",
                                                        },
                                                        {
                                                            "label": "Both",
                                                            "value": "both",
                                                        },
                                                    ],
                                                    value="both",
                                                ),
                                            ],
                                        ),
                                        html.Div(
                                            className="twelve columns",
                                            id="additional-properties",
                                            children=[
                                                html.Hr(),
                                                html.H6("Additional Properties"),
                                                dcc.Checklist(
                                                    id="graph_seconds",
                                                    options=[
                                                        {
                                                            "label": "Graph X Axis in Seconds",
                                                            "value": "show_seconds",
                                                        }
                                                    ],
                                                    value=[],
                                                    labelStyle={
                                                        "display": "inline-block"
                                                    },
                                                ),
                                                dcc.Checklist(
                                                    id="show_grid",
                                                    options=[
                                                        {
                                                            "label": "Show Grid lines",
                                                            "value": "showGridLines",
                                                        }
                                                    ],
                                                    value=["showGridLines"],
                                                    labelStyle={
                                                        "display": "inline-block"
                                                    },
                                                ),
                                                dcc.Checklist(
                                                    id="show_bg",
                                                    options=[
                                                        {
                                                            "label": "Show Background",
                                                            "value": "showBg",
                                                        }
                                                    ],
                                                    value=["showBg"],
                                                    labelStyle={
                                                        "display": "inline-block"
                                                    },
                                                ),
                                                html.Div(
                                                    id="show_mean_graph_title",
                                                    children=[
                                                        html.Label(
                                                            "Group Means Title:"
                                                        ),
                                                        dcc.Input(
                                                            id="mean_graph_title",
                                                            className="seven columns",
                                                        ),
                                                        html.Button(
                                                            children=[
                                                                html.I(
                                                                    n_clicks=0,
                                                                    className="fa fa-chart-area",
                                                                    style={
                                                                        "font-size": "24px"
                                                                    },
                                                                ),
                                                                html.P(
                                                                    "Set",
                                                                    style={
                                                                        "padding-left": "5px"
                                                                    },
                                                                ),
                                                            ],
                                                            id="set_multi_title",
                                                            n_clicks=0,
                                                            className="four columns",
                                                        ),
                                                    ],
                                                    style=set_display(groups_exist),
                                                ),
                                                html.Label("Individual Videos Title:"),
                                                dcc.Input(
                                                    id="ind_graph_title",
                                                    className="seven columns",
                                                ),
                                                html.Button(
                                                    children=[
                                                        html.I(
                                                            n_clicks=0,
                                                            className="fa fa-chart-area",
                                                            style={"font-size": "24px"},
                                                        ),
                                                        html.P(
                                                            "Set",
                                                            style={
                                                                "padding-left": "5px"
                                                            },
                                                        ),
                                                    ],
                                                    id="set_ind_title",
                                                    n_clicks=0,
                                                    className="four columns",
                                                ),
                                                html.Label("Choose Font:"),
                                                dcc.Dropdown(
                                                    id="fonts",
                                                    options=[
                                                        {"label": f, "value": f}
                                                        for f in FONTS
                                                    ],
                                                    value=FONTS[0],
                                                    multi=False,
                                                    clearable=False,
                                                ),
                                                html.Label("Font Size:"),
                                                daq.NumericInput(
                                                    id="font_size",
                                                    value=12,
                                                    min=8,
                                                    max=48,
                                                ),
                                            ],
                                        ),
                                    ],
                                )
                            ],
                        ),
                        dcc.Tab(
                            label="Download Settings",
                            value="download_settings",
                            children=[
                                html.Div(
                                    id="download-settings",
                                    children=[
                                        html.Div(
                                            id="csv_export",
                                            className="twelve columns",
                                            children=[
                                                html.H6("CSV Export"),
                                                html.Label("Enter csv file name:"),
                                                dcc.Input(id="csv_file_name"),
                                                html.P(children=[".csv"]),
                                                html.Button(
                                                    children=[
                                                        html.I(
                                                            n_clicks=0,
                                                            className="fa fa-download",
                                                            style={"font-size": "24px"},
                                                        ),
                                                        html.P(
                                                            "Group Means",
                                                            style={
                                                                "padding-left": "5px"
                                                            },
                                                        ),
                                                    ],
                                                    id="download_group_data",
                                                    n_clicks=0,
                                                    style=set_display(groups_exist),
                                                    className="five columns",
                                                ),
                                                html.Button(
                                                    children=[
                                                        html.I(
                                                            n_clicks=0,
                                                            className="fa fa-download",
                                                            style={"font-size": "24px"},
                                                        ),
                                                        html.P(
                                                            "Videos",
                                                            style={
                                                                "padding-left": "5px"
                                                            },
                                                        ),
                                                    ],
                                                    id="download_data",
                                                    n_clicks=0,
                                                    className="five columns",
                                                ),
                                                html.Div(
                                                    id="comments",
                                                    className="twelve columns",
                                                ),
                                                html.Hr(className="twelve columns"),
                                            ],
                                        ),
                                        html.Div(
                                            id="img_export",
                                            className="twelve columns",
                                            children=[
                                                html.H6(
                                                    "Image Export",
                                                    className="twelve columns",
                                                ),
                                                html.Div(
                                                    id="img-dim",
                                                    className="twelve columns",
                                                    children=[
                                                        html.Label(
                                                            "Enter image dimensions (px):",
                                                            className="twelve columns",
                                                        ),
                                                        daq.NumericInput(
                                                            id="img_height",
                                                            className="three columns",
                                                            value=500,
                                                            min=0,
                                                            max=1000,
                                                        ),
                                                        html.P(
                                                            "x", className="one column"
                                                        ),
                                                        daq.NumericInput(
                                                            id="img_width",
                                                            className="three columns",
                                                            value=700,
                                                            min=0,
                                                            max=1000,
                                                        ),
                                                    ],
                                                ),
                                                html.Label(
                                                    "Enter image name:",
                                                    className="twelve columns",
                                                ),
                                                dcc.Input(id="img_file_name"),
                                                html.P(
                                                    id="img_type", children=[".svg"]
                                                ),
                                                html.Label("File extension:"),
                                                dcc.Dropdown(
                                                    id="img_ext",
                                                    options=[
                                                        {
                                                            "label": "PNG",
                                                            "value": "png",
                                                        },
                                                        {
                                                            "label": "JPEG",
                                                            "value": "jpeg",
                                                        },
                                                        {
                                                            "label": "SVG",
                                                            "value": "svg",
                                                        },
                                                    ],
                                                    value="svg",
                                                ),
                                                html.Button(
                                                    children=[
                                                        html.I(
                                                            n_clicks=0,
                                                            className="fa fa-save",
                                                            style={"font-size": "24px"},
                                                        ),
                                                        html.P(
                                                            "Save Download Settings",
                                                            style={
                                                                "padding-left": "5px"
                                                            },
                                                        ),
                                                    ],
                                                    id="submit_download",
                                                    n_clicks=0,
                                                    type="submit",
                                                ),
                                            ],
                                        ),
                                    ],
                                )
                            ],
                        ),
                    ],
                )
            ],
        ),
        html.Div(
            id="graphs",
            className="eight columns",
            children=[
                html.Div(
                    id="multi_group_container",
                    # style={'display': 'none'},
                    children=[
                        dcc.Graph(
                            id="plot_multi_groups",
                            style={"display": "none"},
                            config=config,
                        )
                    ],
                ),
                dcc.Graph(
                    id="plot_ind_group", config=config, style={"display": "block"}
                ),
            ],
        ),
        html.Div(
            id="credits",
            className="footer",
            children=[
                html.Div(
                    className="six columns",
                    children=[
                        html.P("Created by "),
                        html.A(
                            "Sophia Hwang",
                            href="https://github.com/sophihwang26",
                            target="_blank",
                        ),
                        html.P(" and "),
                        html.A(
                            "Aasiya Islam",
                            href="https://github.com/aasiya-islam",
                            target="_blank",
                        ),
                    ],
                ),
                html.Div(
                    className="six columns",
                    children=[
                        html.P(
                            "Built with Dash",
                            style={
                                "height": "15%",
                                "width": "15%",
                                "float": "right",
                                "position": "relative",
                            },
                        )
                    ],
                ),
            ],
        ),
    ]
)


# Sets config type text
@app.callback(Output("img_type", "children"), [Input("img_ext", "value")])
def update_ext_text(extension):
    return ["." + extension]


# Sets config settings for graph
@app.callback(
    Output("graphs", "children"),
    [Input("submit_download", "n_clicks")],
    [
        State("img_height", "value"),
        State("img_width", "value"),
        State("img_file_name", "value"),
        State("img_ext", "value"),
    ],
)
def update_config(submit, height, width, filename, extension):
    if submit is not None:
        new_config = {
            "toImageButtonOptions": {
                "format": extension,  # one of png, svg, jpeg, webp
                "filename": filename,
                "height": height,
                "width": width,
                "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
            }
        }
        # print('config settings saved')
        return dcc.Graph(id="plot_multi_groups", config=new_config), dcc.Graph(
            id="plot_ind_group", config=new_config
        )
    else:
        return dcc.Graph(id="plot_multi_groups", config=config), dcc.Graph(
            id="plot_ind_group", config=config
        )


# Sets available options in dropdowns based on behavior selected
@app.callback(
    Output("feature_selector", "options"),
    [Input("category_selector", "value"), Input("behaviors", "value")],
)
def set_features_options(selected_category, behavior):
    return [
        {"label": f, "value": f}
        for f in list(all_data_behaviors[behavior][selected_category].keys())
    ]


# Sets list of available videos to plot on graph based on group selected
@app.callback(
    Output("video_numbers", "options"),
    [
        Input("group_selector", "value"),
        Input("category_selector", "value"),
        Input("behaviors", "value"),
        Input("feature_selector", "value"),
    ],
)
def set_group_options(
    selected_group, selected_category, selected_behavior, selected_feature
):
    try:
        return [{"label": v, "value": v} for v in groups_dict[selected_group]]
    except KeyError:
        print("No Groups")
        if "VideoData" in selected_category:
            return [{"label": v, "value": v} for v in H5list_videos]
        else:
            try:
                list_vid = list(
                    all_data_behaviors[selected_behavior][selected_category][
                        selected_feature
                    ].index
                )
                return [{"label": v, "value": v} for v in list_vid]
            except KeyError:
                return []


# Sets default value for Feature
@app.callback(
    Output("feature_selector", "value"), [Input("feature_selector", "options")]
)
def set_feature_value(options):
    try:
        return options[0]["value"]
    except IndexError:
        print("No features found.")
        return ""


# Sets options for category
@app.callback(Output("category_selector", "options"), [Input("behaviors", "value")])
def set_video_option(behavior):
    if behavior == "Attack":
        return [
            {"label": c, "value": c} for c in list_categories if "Severity" not in c
        ]
    else:
        return [{"label": c, "value": c} for c in list_categories]


# Sets default value for Category
@app.callback(
    Output("category_selector", "value"), [Input("category_selector", "options")]
)
def set_category_value(options):
    try:
        return options[0]["value"]
    except IndexError:
        return ""


# Sets default values for videos to plot
@app.callback(Output("video_numbers", "value"), [Input("video_numbers", "options")])
def set_video_option(options):
    try:
        # video_list = [(options[0]['value'])]
        # return video_list
        return [vid["value"] for vid in options]
    except IndexError:
        return ""


# Sets default value for single group to plot
@app.callback(Output("group_selector", "value"), [Input("group_selector", "options")])
def set_video_option(options):
    try:
        return options[0]["value"]
    except IndexError:
        return ""


# Sets visibility of graphing options depending on bar or scatter plot
@app.callback(
    [
        Output("probability-graph-properties", "style"),
        Output("bar-graph-properties", "style"),
        Output("video_mean", "style"),
        Output("group_mean", "style"),
        Output("graph_seconds", "style"),
    ],
    [Input("category_selector", "value")],
)
def graph_options_visibility(category):
    if "VideoData" in category:
        return (
            {"display": "block"},
            {"display": "none"},
            {"display": "block"},
            {"display": "block"},
            {"display": "block"},
        )
    else:
        return (
            {"display": "none"},
            {"display": "block"},
            {"display": "none"},
            {"display": "none"},
            {"display": "none"},
        )


# Return ranges for numeric input given relayoutData and list of all frame
def determine_range(relayoutData, ranges, frames):
    index = 0
    # print(relayoutData)
    for word in json.dumps(relayoutData).split():
        word = word.replace(",", "")
        word = word.replace("}", "")
        # Determine closest frame number given relayoutData ranges
        try:
            number = float(word)
            absolute_different_function = lambda whole_x: abs(whole_x - number)
            closest_value = min(frames, key=absolute_different_function)
            ranges[index] = closest_value
            index += 1
        except ValueError:
            pass
    return ranges


# Sets value of start and end frames in Viewing Range numeric input box
@app.callback(
    [
        Output("ind_start_x", "value"),
        Output("ind_end_x", "value"),
        Output("ind_start_x", "disabled"),
        Output("ind_end_x", "disabled"),
        Output("multi_start_x", "value"),
        Output("multi_end_x", "value"),
        Output("multi_start_x", "disabled"),
        Output("multi_end_x", "disabled"),
        Output("ind_start_sec", "value"),
        Output("ind_end_sec", "value"),
        Output("multi_start_sec", "value"),
        Output("multi_end_sec", "value"),
    ],
    [
        Input("plot_ind_group", "relayoutData"),
        Input("plot_multi_groups", "relayoutData"),
        Input("graph_seconds", "value"),
        Input("feature_selector", "value"),
        Input("submit_multi_range", "n_clicks"),
        Input("submit_ind_range", "n_clicks"),
        Input("reset_multi_range", "n_clicks"),
        Input("reset_ind_range", "n_clicks"),
    ],
    [
        State("behaviors", "value"),
        State("category_selector", "value"),
        State("ind_start_x", "value"),
        State("ind_end_x", "value"),
        State("multi_start_x", "value"),
        State("multi_end_x", "value"),
    ],
)
def set_input_xrange(
    ind_relayout,
    multi_relayout,
    seconds,
    feature,
    submit_multi,
    submit_ind,
    reset_multi,
    reset_ind,
    behavior,
    category,
    prev_start_i,
    prev_end_i,
    prev_start_g,
    prev_end_g,
):
    if "VideoData" not in category:
        return None, None, True, True, None, None, True, True, None, None, None, None
    else:
        x = all_data_behaviors[behavior][category][feature].index
        min_value = 0
        max_value = len(x) - 1
        changed_id = [p["prop_id"] for p in dash.callback_context.triggered][0]
        # print(changed_id)
        disabled = False

        # Initial Load
        if (prev_start_i or prev_end_i or prev_start_g or prev_end_g) is None:
            # print('Previously no Values in numeric input')
            current_indx = [min_value, max_value, 0, 1]
            current_multix = [min_value, max_value, 0, 1]
        elif "reset_ind_range" in changed_id:
            current_indx = [min_value, max_value, 0, 1]
            current_multix = [prev_start_g, prev_end_g, 0, 1]
        elif "reset_multi_range" in changed_id:
            current_indx = [prev_start_i, prev_end_i, 0, 1]
            current_multix = [min_value, max_value, 0, 1]
        # Else, assume the input already has preexisting numeric range values
        else:
            # print('Preexisting value in numeric input')
            current_indx = [prev_start_i, prev_end_i, 0, 1]
            current_multix = [prev_start_g, prev_end_g, 0, 1]

        # If User resized plotly graph, estimate numeric input
        if "graph_seconds" in changed_id:
            # Disable if seconds are enabled
            if len(seconds) > 0:
                disabled = True
        elif "plot_ind_group" in changed_id:
            # print('User resized on plotly')
            current_indx = determine_range(ind_relayout, current_indx, x)
        elif "plot_multi_group" in changed_id:
            current_multix = determine_range(multi_relayout, current_multix, x)

        if len(seconds) > 0:
            disabled = True

        # Compute Seconds
        fps = int(fps_dict[H5list_videos[0]])
        start_i_sec, end_i_sec = round(current_indx[0] / fps, 3), round(
            current_indx[1] / fps, 3
        )
        start_g_sec, end_g_sec = round(current_multix[0] / fps, 3), round(
            current_multix[1] / fps, 3
        )

        return (
            current_indx[0],
            current_indx[1],
            disabled,
            disabled,
            current_multix[0],
            current_multix[1],
            disabled,
            disabled,
            start_i_sec,
            end_i_sec,
            start_g_sec,
            end_g_sec,
        )


# Find and return the total x axis in seconds and the title of x axis, Given list of total frames and behavior name
def show_seconds(start, end, name):
    fps = int(fps_dict[name])
    t = np.arange(start, end + 1 / fps, 1 / fps)
    # print('Seconds Range: ' + str(t[0]) + ' ' + str(t[-1]))
    return t, "Seconds"


# Updates layout of the graph figure,
# given graph figure, plot title, x and y axis titles, boolean show grid, font type, font size
def update_layout(graph, plot_title, x_axis, y_axis, show_grid, font, size, show_bg):
    grid, bg = False, "#ffffff"

    if len(show_grid) > 0:
        grid = True
    if len(show_bg) > 0:
        bg = "#e5ecf6"

    graph["layout"] = go.Layout(
        title={
            "text": plot_title,
            "y": 0.9,
            "x": 0.5,
            "xanchor": "center",
            "yanchor": "top",
        },
        xaxis_title=x_axis,
        yaxis_title=y_axis,
        xaxis={"showgrid": grid},
        yaxis={"showgrid": grid},
        font=dict(family=font, size=size),
        plot_bgcolor=bg,
    )
    return graph


# Returns the start and end x axis range for graph figure,
# given start and end frame given by numeric input, list of all frame numbers
def set_graph_viewing_range(start, end, whole_x):
    try:
        low_number = start
        hi_number = end

        if start > end or start < 0 or end > len(whole_x) - 1:
            pass

        absolute_low_different_function = lambda whole_x: abs(whole_x - low_number)
        closest_low_value = min(whole_x, key=absolute_low_different_function)

        absolute_hi_different_function = lambda whole_x: abs(whole_x - hi_number)
        closest_hi_value = min(whole_x, key=absolute_hi_different_function)

        start_ind = whole_x.index(closest_low_value)
        end_ind = whole_x.index(closest_hi_value)
    # Initial load, numeric input has no value
    except TypeError:
        start_ind = 0
        end_ind = len(whole_x) - 1

    return start_ind, end_ind


@app.callback(
    Output("plot_ind_group", "figure"),
    [
        Input("video_numbers", "value"),
        Input("feature_selector", "value"),
        Input("group_selector", "value"),
        Input("video_mean", "value"),
        Input("update_graph", "n_clicks"),
        Input("ind_group_colors", "colorscale"),
        Input("graph_seconds", "value"),
        Input("show_grid", "value"),
        Input("set_ind_title", "n_clicks"),
        Input("fonts", "value"),
        Input("font_size", "value"),
        Input("show_bg", "value"),
        Input("submit_ind_range", "n_clicks"),
        Input("ind_start_sec", "value"),
        Input("ind_end_sec", "value"),
    ],
    [
        State("behaviors", "value"),
        State("category_selector", "value"),
        State("color_pickers", "children"),
        State("ind_start_x", "value"),
        State("ind_end_x", "value"),
        State("ind_graph_title", "value"),
    ],
)
def update_ind_graph(
    select_vids,
    feature_name,
    group_name,
    show_mean,
    update_colors,
    colorset,
    seconds,
    showgrid,
    update_title,
    font,
    fontsize,
    showbg,
    update_ind_range,
    start_s,
    end_s,
    behavior,
    category,
    colors,
    startx,
    endx,
    custom_title,
):
    fig = go.Figure()
    graph_title = feature_name + " " + group_name
    color_name = {"Mean": colors[len(colors) - 1]["props"]["color"]}
    changed_id = [p["prop_id"] for p in dash.callback_context.triggered][0]

    # Probability Line Graph
    if "Probability" in feature_name:
        x_title = "Frames"
        y_title = "Probability"

        try:
            for v in select_vids:
                # one_line = all_data_behaviors[behavior][category][feature_name][v]
                # x_axis = list(range(all_data_behaviors[behavior][category][feature_name][v].count()))
                one_line = all_data_behaviors[behavior][category][feature_name].eval(v)
                x_axis = list(range(one_line.count()))

                start_x, end_x = set_graph_viewing_range(startx, endx, x_axis)

                # convert x axis to seconds if seconds box is checked
                if len(seconds) > 0:
                    x_axis, x_title = show_seconds(start_s, end_s, v)
                else:
                    # if range in numeric input has been submitted, update the viewing range in plotly'
                    x_axis = list(range(start_x, end_x + 1))

                one_line = one_line[start_x : end_x + 1]

                fig.add_trace(
                    go.Scattergl(
                        x=x_axis,
                        y=one_line,
                        mode="lines",
                        name=v,
                        line=dict(color=random.choice(colorset)),
                    )
                )
            # Calculating Mean Line for Individual Group
            if len(show_mean) > 0:
                if groups_exist:
                    mean_group_name = group_name + "_Mean"
                else:
                    mean_group_name = "Total_Mean"

                x_axis = list(
                    range(
                        all_data_behaviors[behavior][category][feature_name]
                        .eval(mean_group_name)
                        .count()
                    )
                )
                group_mean = all_data_behaviors[behavior][category][feature_name][
                    select_vids
                ].mean(axis=1)
                # print(len(group_mean))

                start_x, end_x = set_graph_viewing_range(startx, endx, x_axis)

                if len(seconds) > 0:
                    x_axis, x_title = show_seconds(start_s, end_s, behavior)
                else:
                    x_axis = list(range(start_x, end_x + 1))

                group_mean = group_mean[start_x : end_x + 1]

                fig.add_trace(
                    go.Scatter(
                        x=x_axis,
                        y=group_mean,
                        mode="lines",
                        name="Group Mean",
                        line=dict(color=color_name["Mean"]),
                    )
                )
        except KeyError:
            print("video not found")
            print(v)
            pass
    # Bar Graph
    else:
        x_title = "Video / Group"
        if "# bout events" in feature_name:
            y_title = "Mean # of Events"
        else:
            y_title = "Mean Time (seconds)"
        try:
            for v in select_vids:
                try:
                    one_bar = [
                        all_data_behaviors[behavior][category][feature_name].loc[v]
                    ]
                except KeyError:
                    one_bar = [0]
                x_axis = [v]

                fig.add_trace(
                    go.Bar(
                        x=x_axis,
                        y=one_bar,
                        text=one_bar,
                        textposition="auto",
                        name=v,
                        marker_color=random.choice(colorset),
                    )
                )
        except KeyError:
            print("Video not found in H5 file.")
            pass

    if "set_ind_title" in changed_id:
        if custom_title is not None:
            graph_title = custom_title

    fig = update_layout(
        fig, graph_title, x_title, y_title, showgrid, font, fontsize, showbg
    )
    return fig


# Input: Pandas dataframe of current group data
# Output: Returns list formatted mean and standard error value for a group, given list of values in a group
def calculate_group_statistics(group_data):
    mean_values = []
    std_error_values = []
    mean_values.append(round(statistics.mean(group_data), 3))
    count = len(group_data)
    std = statistics.stdev(group_data)
    std_error_values.append(round((std / math.sqrt(count)), 3))

    return mean_values, std_error_values


@app.callback(
    Output("plot_multi_groups", "figure"),
    [
        Input("multi_group_selector", "value"),
        Input("feature_selector", "value"),
        Input("group_mean", "value"),
        Input("update_graph", "n_clicks"),
        Input("color_pickers", "children"),
        Input("graph_seconds", "value"),
        Input("show_grid", "value"),
        Input("set_multi_title", "n_clicks"),
        Input("fonts", "value"),
        Input("font_size", "value"),
        Input("show_bg", "value"),
        Input("submit_multi_range", "n_clicks"),
        Input("error_type", "value"),
        Input("ind_start_sec", "value"),
        Input("ind_end_sec", "value"),
    ],
    [
        State("behaviors", "value"),
        State("category_selector", "value"),
        State("multi_start_x", "value"),
        State("multi_end_x", "value"),
        State("mean_graph_title", "value"),
    ],
)
def update_multi_graph(
    list_groups,
    feature_name,
    show_group_mean,
    update_colors,
    colors,
    seconds,
    showgrid,
    update_title,
    font,
    fontsize,
    showbg,
    update_multi_range,
    error,
    start_s,
    end_s,
    behavior,
    category,
    startx,
    endx,
    custom_title,
):
    fig = go.Figure()
    graph_title = feature_name + " - Group Means"
    color_name = {}
    changed_id = [p["prop_id"] for p in dash.callback_context.triggered][0]

    for group in list_groups:
        for c in colors:
            if c["type"] == "ColorPicker" and group in c["props"]["id"]:
                color_name[group] = c["props"]["color"]
    color_name["Mean"] = colors[len(colors) - 1]["props"]["color"]

    if "Probability" in feature_name:
        x_title = "Frames"
        y_title = "Probability"

        for group in list_groups:
            x_axis = list(
                range(
                    all_data_behaviors[behavior][category][feature_name][
                        group + "_Mean"
                    ].count()
                )
            )
            one_mean_line = all_data_behaviors[behavior][category][feature_name][
                group + "_Mean"
            ]
            start_x, end_x = set_graph_viewing_range(startx, endx, x_axis)

            if len(seconds) > 0 and "graph_seconds" in changed_id:
                x_axis, x_title = show_seconds(start_s, end_s, group)
            else:
                x_axis = list(range(start_x, end_x + 1))

            one_mean_line = one_mean_line[start_x : end_x + 1]

            fig.add_trace(
                go.Scattergl(
                    x=x_axis,
                    y=one_mean_line,
                    mode="lines",
                    name=group + " Mean",
                    line=dict(color=color_name[group]),
                )
            )

        if len(show_group_mean) > 0:
            x_axis = list(
                range(
                    all_data_behaviors[behavior][category][feature_name][
                        "Total_Mean"
                    ].count()
                )
            )
            total_mean_line = all_data_behaviors[behavior][category][feature_name][
                "Total_Mean"
            ]
            start_x, end_x = set_graph_viewing_range(startx, endx, x_axis)

            if len(seconds) > 0:
                x_axis, x_title = show_seconds(startx, endx, behavior)
            else:
                x_axis = list(range(start_x, end_x + 1))

            total_mean_line = total_mean_line[start_x : end_x + 1]

            fig.add_trace(
                go.Scatter(
                    x=x_axis,
                    y=total_mean_line,
                    mode="lines",
                    name="Overall Mean",
                    line=dict(color=color_name["Mean"]),
                )
            )
    # Bar Graph
    else:
        x_title = "Video"
        if "# bout events" in feature_name:
            y_title = "Mean # of Events"
        else:
            y_title = "Mean Time (seconds)"
        try:
            for group in list_groups:
                try:
                    group_values = (
                        all_data_behaviors[behavior][category][feature_name]
                        .loc[groups_dict[group]]
                        .tolist()
                    )
                    mean_values, std_error_values = calculate_group_statistics(
                        group_values
                    )
                except KeyError:
                    mean_values, std_error_values = [0], [0]
                x_axis = [group]

                # print(group_values)

                if "both" in error:
                    symmetry = True
                    array = std_error_values
                    array_minus = []
                elif "above" in error:
                    symmetry = False
                    array = std_error_values
                    array_minus = [0]
                else:
                    symmetry = False
                    array = [0]
                    array_minus = std_error_values

                fig.add_trace(
                    go.Bar(
                        x=x_axis,
                        y=mean_values,
                        text=mean_values,
                        textposition="auto",
                        error_y=dict(
                            type="data",
                            symmetric=symmetry,
                            array=array,
                            arrayminus=array_minus,
                        ),
                        hovertext=std_error_values,
                        name=group,
                        marker_color=color_name[group],
                    )
                )
        except KeyError:
            print("Group not found.")
            pass

    changed_id = [p["prop_id"] for p in dash.callback_context.triggered][0]

    if "set_multi_title" in changed_id:
        if custom_title is not None:
            graph_title = custom_title

    fig = update_layout(
        fig, graph_title, x_title, y_title, showgrid, font, fontsize, showbg
    )

    return fig


# Input: Graph figure, most recently changed property
# Output: Edited Pandas dataframe, title of output csv
def load_results(graph_fig, prop_id):
    df = pd.DataFrame()
    csv_title = graph_fig["layout"]["title"]["text"]

    if "Probability" in csv_title:
        for line in graph_fig["data"]:
            # print(line)
            s = pd.Series(line["y"])
            # print(s)
            df = pd.concat([df, s.rename(line["name"])], axis=1)
        index_name = graph_fig["layout"]["xaxis"]["title"]["text"]

        df[index_name] = graph_fig["data"][0]["x"]
        df = df.set_index(index_name)

    else:
        groups = []
        data = []
        error = []

        for bar in graph_fig["data"]:
            groups += bar["x"]
            data += bar["y"]
            if "download_group_data" in prop_id:
                error += bar["error_y"]["array"]

        df["Groups"] = groups
        df[csv_title] = data
        if "download_group_data" in prop_id:
            df["Error"] = error

        df = df.set_index("Groups")
    return df, csv_title


# Downloads data .csv to project folder
@app.callback(
    Output("comments", "children"),
    [Input("download_data", "n_clicks"), Input("download_group_data", "n_clicks")],
    [
        State("plot_ind_group", "figure"),
        State("plot_multi_groups", "figure"),
        State("csv_file_name", "value"),
    ],
)
def download(ind_click, multi_click, ind_figure, multi_figure, file_name):
    changed_id = [p["prop_id"] for p in dash.callback_context.triggered][0]

    if "download_data" in changed_id:
        figure = ind_figure

        click = ind_click
    else:
        figure = multi_figure
        click = multi_click

    if click > 0:
        download_df, str_name = load_results(figure, changed_id)
        # print(download_df)
        if file_name is not None:
            str_name = file_name

        # Set csv save path to csv folder
        csvSavePath = get_download_path()
        csvSavePath = os.path.join(csvSavePath, str_name + ".csv")
        download_df.to_csv(csvSavePath)
        # print('Downloaded csv as ' + str_name + '.csv')
        return ["Saved " + str_name + ".csv to Downloads"]
    else:
        return []


if __name__ == "__main__":
    app.run_server(debug=False)



__author__ = "Simon Nilsson / Florian Duclot"

import glob
import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.third_party_label_appenders.tools import is_new_boris_version
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.errors import (ThirdPartyAnnotationEventCountError,
                                ThirdPartyAnnotationOverlapError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import (
    ThirdPartyAnnotationsInvalidFileFormatWarning,
    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning)


class BorisAppender(ConfigReader):
    """
    Append BORIS human annotations onto featurized pose-estimation data.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: path to folder holding BORIS data files is CSV format

    .. note::
       `Third-party import tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.
       `Example I BORIS input file <https://github.com/sgoldenlab/simba/blob/master/misc/boris_example.csv>`_.
       `Example II BORIS input file <https://github.com/sgoldenlab/simba/blob/master/misc/boris_new_example.csv>`_.

    .. image:: _static/img/boris.png
       :width: 200
       :align: center

    Examples
    ----------
    >>> boris_appender = BorisAppender(config_path='MyProjectConfigPath', data_dir=r'BorisDataFolder')
    >>> boris_appender.create_boris_master_file()
    >>> boris_appender.run()

    References
    ----------

    .. [1] `Behavioral Observation Research Interactive Software (BORIS) user guide <https://boris.readthedocs.io/en/latest/#>`__.
    """

    def __init__(self, config_path: str, data_dir: str):
        super().__init__(config_path=config_path)
        self.boris_dir = data_dir
        self.boris_files_found = glob.glob(self.boris_dir + "/*.csv")
        check_if_dir_exists(data_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.boris_files_found,
            error_msg=f"SIMBA ERROR: 0 BORIS CSV files found in {data_dir} directory",
        )
        print(f"Processing BORIS for {str(len(self.feature_file_paths))} file(s)...")

    def create_boris_master_file(self):
        """
        Method to create concatenated dataframe of BORIS annotations.

        Returns
        -------
        Attribute: pd.Dataframe
            master_boris_df
        """
        self.master_boris_df_list = []
        for file_cnt, file_path in enumerate(self.boris_files_found):
            try:
                _, video_name, _ = get_fn_ext(file_path)
                boris_df = pd.read_csv(file_path)
                if not is_new_boris_version(boris_df):
                    index = boris_df[boris_df["Observation id"] == "Time"].index.values
                    boris_df = pd.read_csv(file_path, skiprows=range(0, int(index + 1)))
                    boris_df = boris_df.loc[
                        :, ~boris_df.columns.str.contains("^Unnamed")
                    ]
                    boris_df.drop(
                        ["Behavioral category", "Comment", "Subject"],
                        axis=1,
                        inplace=True,
                    )
                else:
                    target_cols = [
                        "Time",
                        "Media file path",
                        "Total length",
                        "FPS",
                        "Behavior",
                        "Status",
                    ]
                    boris_df.rename(
                        columns={
                            "Media file name": "Media file path",
                            "Media duration (s)": "Total length",
                            "Behavior type": "Status",
                        },
                        inplace=True,
                    )
                    boris_df = boris_df.reindex(columns=target_cols)
                _, video_base_name, _ = get_fn_ext(boris_df.loc[0, "Media file path"])
                boris_df["Media file path"] = video_base_name
                self.master_boris_df_list.append(boris_df)
            except Exception as e:
                print(e.args)
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="BORIS", file_path=file_path
                )
        self.master_boris_df = pd.concat(self.master_boris_df_list, axis=0).reset_index(
            drop=True
        )
        print(
            "Found {} annotated behaviors in {} files with {} directory".format(
                str(len(self.master_boris_df["Behavior"].unique())),
                len(self.boris_files_found),
                self.boris_dir,
            )
        )
        print(
            "The following behavior annotations where detected in the boris directory:"
        )
        for behavior in self.master_boris_df["Behavior"].unique():
            print(behavior)

    def __check_non_overlapping_annotations(self):
        shifted_annotations = deepcopy(self.clf_annotations)
        shifted_annotations["START"] = self.clf_annotations["START"].shift(-1)
        shifted_annotations = shifted_annotations.head(-1)
        return shifted_annotations.query("START < STOP")

    def run(self):
        """
        Method to append BORIS annotations created in :meth:`~simba.BorisAppender.create_boris_master_file` to the
        featurized pose-estimation data in the SimBA project. Results (parquets' or CSVs) are saved within the the
        project_folder/csv/targets_inserted directory of the SimBA project.
        """
        for file_cnt, file_path in enumerate(self.feature_file_paths):
            _, self.video_name, _ = get_fn_ext(file_path)
            print("Appending BORIS annotations to {} ...".format(self.video_name))
            data_df = read_df(file_path, self.file_type)
            self.out_df = deepcopy(data_df)
            vid_annotations = self.master_boris_df.loc[
                self.master_boris_df["Media file path"] == self.video_name
            ]
            vid_annotation_starts = vid_annotations[(vid_annotations.Status == "START")]
            vid_annotation_stops = vid_annotations[(vid_annotations.Status == "STOP")]
            vid_annotations = (
                pd.concat(
                    [vid_annotation_starts, vid_annotation_stops],
                    axis=0,
                    join="inner",
                    copy=True,
                )
                .sort_index()
                .reset_index(drop=True)
            )
            vid_annotations = vid_annotations[
                vid_annotations["Behavior"].isin(self.clf_names)
            ]
            if len(vid_annotations) == 0:
                print(
                    "SIMBA WARNING: No BORIS annotations detected for SimBA classifier(s) named {} for video {}".format(
                        str(self.clf_names), self.video_name
                    )
                )
                continue
            video_fps = vid_annotations["FPS"].values[0]
            for clf in self.clf_names:
                self.clf = clf
                clf_annotations = vid_annotations[(vid_annotations["Behavior"] == clf)]
                clf_annotations_start = clf_annotations[
                    clf_annotations["Status"] == "START"
                ].reset_index(drop=True)
                clf_annotations_stop = clf_annotations[
                    clf_annotations["Status"] == "STOP"
                ].reset_index(drop=True)
                if len(clf_annotations_start) != len(clf_annotations_stop):
                    raise ThirdPartyAnnotationEventCountError(
                        video_name=self.video_name,
                        clf_name=self.clf,
                        start_event_cnt=len(clf_annotations_start),
                        stop_event_cnt=len(clf_annotations_stop),
                    )
                self.clf_annotations = (
                    clf_annotations_start["Time"]
                    .to_frame()
                    .rename(columns={"Time": "START"})
                )
                self.clf_annotations["STOP"] = clf_annotations_stop["Time"]
                self.clf_annotations = self.clf_annotations.apply(pd.to_numeric)
                results = self.__check_non_overlapping_annotations()
                if len(results) > 0:
                    raise ThirdPartyAnnotationOverlapError(
                        video_name=self.video_name, clf_name=self.clf
                    )
                self.clf_annotations["START_FRAME"] = (
                    self.clf_annotations["START"] * video_fps
                ).astype(int)
                self.clf_annotations["END_FRAME"] = (
                    self.clf_annotations["STOP"] * video_fps
                ).astype(int)
                if len(self.clf_annotations) == 0:
                    self.out_df[clf] = 0
                    print(
                        f"SIMBA WARNING: No BORIS annotation detected for video {self.video_name} and behavior {clf}. SimBA will set all frame annotations as absent."
                    )
                    continue
                annotations_idx = list(
                    self.clf_annotations.apply(
                        lambda x: list(
                            range(int(x["START_FRAME"]), int(x["END_FRAME"]) + 1)
                        ),
                        1,
                    )
                )
                annotations_idx = [x for xs in annotations_idx for x in xs]
                idx_difference = list(set(annotations_idx) - set(self.out_df.index))
                if len(idx_difference) > 0:
                    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
                        video_name=self.video_name,
                        clf_name=clf,
                        frm_cnt=self.out_df.index[-1],
                        first_error_frm=idx_difference[0],
                        ambiguous_cnt=len(idx_difference),
                    )
                    annotations_idx = [
                        x for x in annotations_idx if x not in idx_difference
                    ]
                self.out_df[clf] = 0
                self.out_df.loc[annotations_idx, clf] = 1
            self.__save_boris_annotations()
        self.timer.stop_timer()
        stdout_success(
            msg="BORIS annotations appended to dataset and saved in project_folder/csv/targets_inserted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save_boris_annotations(self):
        self.save_path = os.path.join(
            self.targets_folder, self.video_name + "." + self.file_type
        )
        write_df(self.out_df, self.file_type, self.save_path)
        print("Saved BORIS annotations for video {}...".format(self.video_name))


# test = BorisAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      data_dir='/Users/simon/Downloads/FIXED')
# test.create_boris_master_file()
# test.run()

# test = BorisAppender(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', boris_folder=r'/Users/simon/Desktop/troubleshooting/train_model_project/boris_import')
# test.create_boris_master_file()
# test.append_boris()

# test = BorisAppender(config_path='/Users/simon/Desktop/envs/marcel_boris/project_folder/project_config.ini', boris_folder=r'/Users/simon/Desktop/envs/marcel_boris/BORIS_data')
# test.create_boris_master_file()
# test.append_boris()

# test = BorisAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      boris_folder='/Users/simon/Downloads/FIXED')
# test.create_boris_master_file()
# test.append_boris()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.pose_processors.reorganize_keypoint import KeypointReorganizer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        FolderSelect, hxtScrollbar)
from simba.utils.enums import Formats, Keys, Links


class PoseReorganizerPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="RE-ORGANIZE POSE_ESTIMATION DATA", size=(500, 800)
        )
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.data_folder = FolderSelect(settings_frm, "DATA FOLDER: ", lblwidth="10")
        self.pose_tool_dropdown = DropDownMenu(
            settings_frm, "Tracking tool", ["DLC", "maDLC"], "10"
        )
        self.pose_tool_dropdown.setChoices("DLC")
        self.file_format = DropDownMenu(
            settings_frm, "FILE TYPE: ", ["csv", "h5"], "10"
        )
        self.file_format.setChoices("csv")
        confirm_btn = Button(
            settings_frm, text="Confirm", command=lambda: self.confirm()
        )
        settings_frm.grid(row=0, sticky=NW)
        self.data_folder.grid(row=0, sticky=NW, columnspan=3)
        self.pose_tool_dropdown.grid(row=1, sticky=NW)
        self.file_format.grid(row=2, sticky=NW)
        confirm_btn.grid(row=2, column=1, sticky=NW)

    def confirm(self):
        if hasattr(self, "table"):
            self.table.destroy()

        self.keypoint_reorganizer = KeypointReorganizer(
            data_folder=self.data_folder.folder_path,
            pose_tool=self.pose_tool_dropdown.getChoices(),
            file_format=self.file_format.getChoices(),
        )
        self.table = LabelFrame(
            self.main_frm,
            text="SET NEW ORDER",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.current_order = LabelFrame(self.table, text="CURRENT ORDER:")
        self.new_order = LabelFrame(self.table, text="NEW ORDER:")

        self.table.grid(row=1, sticky=W, pady=10)
        self.current_order.grid(row=0, column=0, sticky=NW, pady=5)
        self.new_order.grid(row=0, column=1, sticky=NW, padx=5, pady=5)
        idx1, idx2, oldanimallist, oldbplist, self.newanimallist, self.newbplist = (
            [0] * len(self.keypoint_reorganizer.bp_list) for i in range(6)
        )

        if self.keypoint_reorganizer.animal_list:
            animal_list_reduced = list(set(self.keypoint_reorganizer.animal_list))
            self.pose_tool = "maDLC"
            for i in range(len(self.keypoint_reorganizer.bp_list)):
                idx1[i] = Label(self.current_order, text=str(i + 1) + ".")
                oldanimallist[i] = Label(
                    self.current_order,
                    text=str(self.keypoint_reorganizer.animal_list[i]),
                )
                oldbplist[i] = Label(
                    self.current_order, text=str(self.keypoint_reorganizer.bp_list[i])
                )
                idx1[i].grid(row=i, column=0, sticky=W)
                oldanimallist[i].grid(row=i, column=1, sticky=W, ipady=5)
                oldbplist[i].grid(row=i, column=2, sticky=W, ipady=5)
                idx2[i] = Label(self.new_order, text=str(i + 1) + ".")
                self.newanimallist[i] = DropDownMenu(
                    self.new_order, " ", animal_list_reduced, "10"
                )
                self.newbplist[i] = DropDownMenu(
                    self.new_order, " ", self.keypoint_reorganizer.bp_list, "10"
                )
                self.newanimallist[i].setChoices(
                    self.keypoint_reorganizer.animal_list[i]
                )
                self.newbplist[i].setChoices(self.keypoint_reorganizer.bp_list[i])
                idx2[i].grid(row=i, column=0, sticky=W)
                self.newanimallist[i].grid(row=i, column=1, sticky=W)
                self.newbplist[i].grid(row=i, column=2, sticky=W)

        else:
            self.pose_tool = "DLC"
            for i in range(len(self.keypoint_reorganizer.bp_list)):
                idx1[i] = Label(self.current_order, text=str(i + 1) + ".")
                oldbplist[i] = Label(
                    self.current_order, text=str(self.keypoint_reorganizer.bp_list[i])
                )
                idx1[i].grid(row=i, column=0, sticky=W, ipady=5)
                oldbplist[i].grid(row=i, column=2, sticky=W, ipady=5)
                idx2[i] = Label(self.new_order, text=str(i + 1) + ".")
                self.newbplist[i] = StringVar()
                oldanimallist[i] = OptionMenu(
                    self.new_order,
                    self.newbplist[i],
                    *self.keypoint_reorganizer.bp_list
                )
                self.newbplist[i].set(self.keypoint_reorganizer.bp_list[i])
                idx2[i].grid(row=i, column=0, sticky=W)
                oldanimallist[i].grid(row=i, column=1, sticky=W)

        button_run = Button(
            self.table,
            text="Run re-organization",
            command=lambda: self.run_reorganization(),
        )
        button_run.grid(row=2, column=1, sticky=W)

    def run_reorganization(self):
        if self.pose_tool == "DLC":
            new_bp_list = []
            for curr_choice in self.newbplist:
                new_bp_list.append(curr_choice.get())
            self.keypoint_reorganizer.run(animal_list=None, bp_lst=new_bp_list)

        if self.pose_tool == "maDLC":
            new_bp_list, new_animal_list = [], []
            for curr_animal, curr_bp in zip(self.newanimallist, self.newbplist):
                new_bp_list.append(curr_bp.getChoices())
                new_animal_list.append(curr_animal.getChoices())
            self.keypoint_reorganizer.run(
                animal_list=new_animal_list, bp_lst=new_bp_list
            )
